/****************************************************************************
** Demonstration of playing an Audio CD                                    **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include "conio.h"
#include "dos.h"

//Structure used to transfer commands to the CDROM driver
struct CommandStructure {
  unsigned char Length;               //Length of parameter block
  unsigned char SUC;                  //SubUnitCode of CDROM drive,
                                      //MSCDEX fills this in for us
  unsigned char FNr;                  //Function number
  unsigned int  SW;                   //Driver returns: Status word
  unsigned char Unused [9];           //Unused
  unsigned int  PBOfs;                //Pointer to control block containing
  unsigned int  PBSeg;                //data for the specific function call
} Command;

//Structure to receive information on the audio CD
struct CDInfoStructure {
  unsigned char SubFNr;        //CDROM driver subfunction
  unsigned char First;         //First track
  unsigned char Last;          //Last track
  unsigned long LeadOut;       //Location of the end of the last track
};

//Structure used to receive info on a specific track on the CD
struct TrackInfoStructure {
  unsigned char SubFNr;               //SubFunction number
  unsigned char TrackNr;              //Track number
  unsigned long Address;              //Driver returns: Sector address
  unsigned char Status;               //Driver returns: Track status
};

//Structure used to start playing an audio track
struct PlayStructure {
  unsigned char Length;               //Length of parameter block
  unsigned char SUC;                  //SubUnitCode of CDROM drive,
                                      //MSCDEX fills this in for us
  unsigned char FNr;                  //Function number
  unsigned int  SW;                   //Driver returns: Status word
  unsigned char Unused [8];           //Unused
  unsigned char Mode;                 //Addressing mode (Using 1, Red Book)
  unsigned long Start;                //Starting sector of the track
  unsigned long Sectors;              //Amount of sectors to play
  //Bear in mind that one 'sector' equals one 75th of a second
};

//Structure used to stop playing an audio track
struct StopStructure {
  unsigned char Length;               //Length of parameter block
  unsigned char SUC;                  //SubUnitCode of CDROM drive,
                                      //MSCDEX fills this in for us
  unsigned char FNr;                  //Function number
  unsigned int  SW;                   //Driver returns: Status word
};

  unsigned char Number_Of_Players;    //The amount of players installed
  unsigned int  Drive_Letter;         //The drive letter of the first player
  unsigned char First_Track;          //The first track on the current disc
  unsigned char Last_Track;           //The last track on the current disc
  unsigned long LeadOut_Track;        //The end of the current disc
  unsigned int  StatusWord;           //The result of a function call is
                                      //stored here

  unsigned char TrackToPlay;          //Stores the track the user wants to
                                      //play
  char          TrackToPlayASCII[3];  //Used for keyboard input of the above
                                      //value;
  unsigned char Done;                 //Global indicator

/**************************************************************************
** Check whether MSCDEX is installed and controlling at least one player **
**************************************************************************/
unsigned char Check_MSCDEX ()
{
  union REGS Regs;

  Regs.x.ax = 0x1500; //Function 0 - Get number of CDROM players installed
  Regs.x.bx = 0;
  int86 (0x2F, &Regs, &Regs);
  if (Regs.x.bx) {
    //If BX <> 0 then at least one player has been installed
    //Store the number of CDROM players
    Number_Of_Players = Regs.x.bx;
    //Store the drive indication of the first player
    Drive_Letter      = Regs.x.cx;
    /*Note: this programme only controls the first CDROM player installed,
            in theory it is possible to control more than one player,
            however, it is unlikely that someone either *has* more than
            one player or that a programmer wishes to play music through
            more than one player*/
    //Return true
    return (1);
  } else {
    //If BX = 0 then no players have been installed
    //Return false
    return (0);
  }
}

/**************************************************************************
** Converts RedBook sector notation to HSG sector number.                **
***************************************************************************
**  When calculating the amount of sectors to be played, it's nice to    **
**  have a start and end offset so we can just subtract them.            **
**  This expects a Red Book sector address in the form of                **
**    MMMMSSFF (minutes * 10000h + seconds * 100h + frames)              **
**  and returns the LSN (logical sector number)                          **
**************************************************************************/
unsigned long HSG (unsigned long RedBook)
{
  unsigned char Frames;
  unsigned char Seconds;
  unsigned int  Minutes;

  Frames  = (RedBook & 0xFF);
  Seconds = (RedBook & 0xFF00) >> 8;
  Minutes =  RedBook >> 16;
  return (Frames + (long)Seconds * 75 + (long)Minutes * 60 * 75);
}

/**************************************************************************
** Call MSCDEX which in turn calls the CDROM driver.                     **
***************************************************************************
**  This expects a control structure for a parameter.                    **
**************************************************************************/
void Call_CDROM_Driver (DrivParms)
{
  union REGS Regs;
  struct SREGS SRegs;

  //Set up for call to MSCDEX, which in turn calls the driver
  Regs.x.ax = 0x1510;               //Function 10 - call CDROM driver
  Regs.x.cx = Drive_Letter;         //Drive letter for first player
  SRegs.es  = FP_SEG(&DrivParms);   //Pointer to Parameterblock
  Regs.x.bx = FP_OFF(&DrivParms);
  int86x (0x2F, &Regs, &Regs, &SRegs);   //Call MSCDEX
}

/**************************************************************************
** Get information regarding the current CD.                             **
***************************************************************************
**  That information consists of                                         **
**    - The number of the first track on the CD                          **
**    - The number of the last track on the CD                           **
**    - The total playing length of the CD                               **
**************************************************************************/
void GetAudioCDInfo ()
{
  struct CDInfoStructure AudioCDInfo;

  Command.FNr      = 3;                //Function 3 - Direct Read
  Command.PBOfs    = FP_OFF(&AudioCDInfo);
  Command.PBSeg    = FP_SEG(&AudioCDInfo);
  Command.Length   = 17;

  AudioCDInfo.SubFNr = 0xA;            //SubFunction 10 - get audio CD info
  Call_CDROM_Driver (Command);         //Call CDROM Driver
  First_Track   = AudioCDInfo.First;   //Store first track number
  Last_Track    = AudioCDInfo.Last;    //Store last track number
  LeadOut_Track = AudioCDInfo.LeadOut; //Store end of CD
  StatusWord    = Command.SW;
}

/**************************************************************************
** Starts playback of a specific track on the CD.                        **
**************************************************************************/
void PlayTrack (unsigned char Track)
{
  struct TrackInfoStructure AudioTrackInfo;
  struct PlayStructure      Play;

  unsigned long             StartOffSet;
  unsigned long             StopOffSet;
  unsigned long             SectorsToPlay;

  Command.FNr      = 3;                //Function 3 - Direct Read
  Command.PBOfs    = FP_OFF(&AudioTrackInfo);
  Command.PBSeg    = FP_SEG(&AudioTrackInfo);
  Command.Length   = 17;

  AudioTrackInfo.SubFNr  = 0xB;   //Function 11 - Get audio track info
  AudioTrackInfo.TrackNr = Track; //Number of the track
  Call_CDROM_Driver(Command);

  //Check to see whether the track is indeed an audio track
  if ((AudioTrackInfo.Status & 0x40) == 0x40) {
    //If not, display error message and don't try to start playing
    printf ("Can't play data track!\n");
    return;
  }

  StartOffSet = AudioTrackInfo.Address; //Start of the track

  //End of the track
  if (Track == Last_Track) {
    //If it is the last, track, the end of the track is the LeadOut
    StopOffSet = LeadOut_Track;
  } else {
    //Otherwise, the end of this track is the start of the following track
    Command.FNr      = 3;               //Function 3 - Direct Read
    Command.PBOfs    = FP_OFF(&AudioTrackInfo);
    Command.PBSeg    = FP_SEG(&AudioTrackInfo);
    Command.Length   = 17;

    AudioTrackInfo.SubFNr  = 0xB;       //Function 11 - Get audio track info
    AudioTrackInfo.TrackNr = Track + 1; //Number of the next track
    Call_CDROM_Driver(Command);
    StopOffSet = AudioTrackInfo.Address;
  }

  //Calculate the amount of sectors to be played
  SectorsToPlay = HSG (StopOffSet) - HSG (StartOffSet);

  //All right! That set up the start and end offsets of the track. Now we
  //start the actual playing.
  Play.FNr     = 0x84;                   //Function 84h - play audio track
  Play.Mode    = 1;                      //Select Red Book addressing mode
  Play.Start   = StartOffSet;            //Start of track
  Play.Sectors = SectorsToPlay;          //Amount of sectors to play
  Play.Length  = 21;
  Call_CDROM_Driver (Play);              //Call MSCDEX
  StatusWord = Play.SW;                  //Return status word
}

/**************************************************************************
** Stops audio playback.                                                 **
**************************************************************************/
void StopPlaying ()
{
  struct StopStructure Stop;

  Stop.FNr     = 0x85;        //Function 85h - stop playback
  Stop.Length  = 5;
  Call_CDROM_Driver (Stop);   //Call MSCDEX
  StatusWord = Stop.SW;       //Return status word
}

/**************************************************************************
** Check the statusword to see if any errors have occurred.              **
***************************************************************************
**  This should be called after any function call                        **
**  Returns the number of the error                                      **
**************************************************************************/
unsigned char CheckForErrors ()
{
  //Check to see whether there was an error
  if (StatusWord & 0x8000 > 0) {
    switch (StatusWord & 0xFF) {
      case  0 : printf ("Can't write.\n"); break;
      case  1 : printf ("Unknown player.\n"); break;
      case  2 : printf ("No CD in player.\n"); break;
      case  3 : printf ("Unknown function.\n"); break;
      case  4 : printf ("CRC Error.\n"); break;
      case  5 : printf ("Error in parameterblock.\n"); break;
      case  6 : printf ("Seek fault.\n"); break;
      case  7 : printf ("Unknown mediatype.\n"); break;
      case  8 : printf ("Sector not found.\n"); break;
      case  9 : printf ("Reserved.\n"); break;
      case 10 : printf ("Write fault.\n"); break;
      case 11 : printf ("Read fault.\n"); break;
      case 12 : printf ("General failure error.\n"); break;
      case 13 : printf ("Reserved.\n"); break;
      case 14 : printf ("Reserved.\n"); break;
      case 15 : printf ("Invalid CD Change.\n"); break;
     }
    return (StatusWord & 0xFF);
  } else {
    return (0);
  }
}

void main ()
{
  clrscr ();
  //Check to see whether MSCDEX and at least one CDROM player are installed
  if ( !Check_MSCDEX() ) {
    //If not, display error message and end programme
    printf ("MSCDEX not installed or no CDROM drive available.\n");
    return;
  } else {
    //If installed, display letter of first CDROM drive
    printf ("CDROM installed as drive %c:\n", Drive_Letter + 65);
  }

  do {
    //Get information on the number of tracks on the CD
    GetAudioCDInfo ();

    if ( !CheckForErrors() ) {
      //Ask the user which track to play
      printf ("\n");
      printf ("Tracks on CD are %d through %d\n" , First_Track , Last_Track);
      printf ("Total playing time: %dm %d.%ds\n",
         LeadOut_Track >> 16,
        (LeadOut_Track & 0xFF00) >> 8,
        (100*LeadOut_Track & 0xFF) / 75);
      printf ("Choose 0 to end programme.\n");
      printf ("Please choose a track to play: ");
      gets (TrackToPlayASCII);
      TrackToPlay = atoi (TrackToPlayASCII);

      //Stop audio playback
      StopPlaying ();
      CheckForErrors ();
    }

    //Check user input
    if (!TrackToPlay) {
      //If user has chosen to quit, set quit flag
      Done = 1;
    } else {
      //otherwise, play the chosen track
      PlayTrack (TrackToPlay);
      CheckForErrors ();
    }

  } while (!Done);
}